import datetime
import uuid

from pyramid.authentication import BasicAuthAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.config import Configurator
from pyramid.renderers import JSON
from pyramid.security import NO_PERMISSION_REQUIRED

from channelstream import patched_json as json
from channelstream.wsgi_views.wsgi_security import APIFactory


def datetime_adapter(obj, request):
    return obj.isoformat()


def uuid_adapter(obj, request):
    return str(obj)


def make_app(server_config):
    config = Configurator(
        settings=server_config, root_factory=APIFactory, default_permission="access"
    )
    config.include("pyramid_jinja2")

    def check_function(username, password, request):
        if (
            password == server_config["admin_secret"]
            and username == server_config["admin_user"]
        ):
            return "admin", username
        return None

    authn_policy = BasicAuthAuthenticationPolicy(check_function, realm="ChannelStream")
    authz_policy = ACLAuthorizationPolicy()

    config.set_authentication_policy(authn_policy)
    config.set_authorization_policy(authz_policy)

    json_renderer = JSON(serializer=json.dumps, indent=4)
    json_renderer.add_adapter(datetime.datetime, datetime_adapter)
    json_renderer.add_adapter(uuid.UUID, uuid_adapter)
    config.add_renderer("json", json_renderer)

    config.add_subscriber(
        "channelstream.subscribers.handle_new_request", "pyramid.events.NewRequest"
    )
    config.add_request_method("channelstream.utils.handle_cors", "handle_cors")
    config.include("channelstream.wsgi_views")
    config.scan("channelstream.wsgi_views.server")
    config.scan("channelstream.wsgi_views.error_handlers")
    config.scan("channelstream.events")

    config.include("pyramid_apispec.views")
    config.pyramid_apispec_add_explorer(
        spec_route_name="openapi_spec",
        script_generator="channelstream.utils:swagger_ui_script_template",
        permission=NO_PERMISSION_REQUIRED,
    )
    app = config.make_wsgi_app()
    return app



import logging
from pyramid.view import exception_view_config

log = logging.getLogger(__name__)


@exception_view_config(context="marshmallow.ValidationError", renderer="json")
def marshmallow_invalid_data(context, request):
    request.response.status = 422
    log.error("Request had incorrect payload")
    return context.messages


@exception_view_config(context="itsdangerous.BadTimeSignature", renderer="json")
@exception_view_config(context="itsdangerous.BadSignature", renderer="json")
def itsdangerous_signer_error(context, request):
    request.response.status = 401
    log.error("Request had incorrect signature")
    return {"request": "Bad Signature"}



import logging

from channelstream.channel import Channel
from channelstream.connection import Connection
from channelstream.server_state import get_state
from channelstream.user import User

log = logging.getLogger(__name__)


def connect(
    username=None,
    fresh_user_state=None,
    state_public_keys=None,
    update_user_state=None,
    conn_id=None,
    channels=None,
    channel_configs=None,
):
    """

    :param username:
    :param fresh_user_state:
    :param state_public_keys:
    :param update_user_state:
    :param conn_id:
    :param channels:
    :param channel_configs:
    :return:
    """
    server_state = get_state()
    with server_state.lock:
        if username not in server_state.users:
            user = User(username)
            user.state_from_dict(fresh_user_state)
            server_state.users[username] = user
        else:
            user = server_state.users[username]
        if state_public_keys is not None:
            user.state_public_keys = state_public_keys

        user.state_from_dict(update_user_state)
        connection = Connection(username, conn_id)
        if connection.id not in server_state.connections:
            server_state.connections[connection.id] = connection
        user.add_connection(connection)
        for channel_name in channels:
            # user gets assigned to a channel
            if channel_name not in server_state.channels:
                channel = Channel(
                    channel_name, channel_config=channel_configs.get(channel_name)
                )
                server_state.channels[channel_name] = channel
            server_state.channels[channel_name].add_connection(connection)
        log.info("connecting %s with uuid %s" % (username, connection.id))
        return connection, user


def subscribe(connection=None, channels=None, channel_configs=None):
    """

    :param connection:
    :param channels:
    :param channel_configs:
    :return:
    """
    server_state = get_state()
    user = server_state.users.get(connection.username)
    subscribed_to = []
    with server_state.lock:
        if user:
            for channel_name in channels:
                if channel_name not in server_state.channels:
                    channel = Channel(
                        channel_name, channel_config=channel_configs.get(channel_name)
                    )
                    server_state.channels[channel_name] = channel
                is_found = server_state.channels[channel_name].add_connection(
                    connection
                )
                if is_found:
                    subscribed_to.append(channel_name)
    return subscribed_to


def unsubscribe(connection=None, unsubscribe_channels=None):
    """

    :param connection:
    :param unsubscribe_channels:
    :return:
    """
    server_state = get_state()
    user = server_state.users.get(connection.username)
    unsubscribed_from = []
    with server_state.lock:
        if user:
            for channel_name in unsubscribe_channels:
                if channel_name in server_state.channels:
                    is_found = server_state.channels[channel_name].remove_connection(
                        connection
                    )
                    if is_found:
                        unsubscribed_from.append(channel_name)
    return unsubscribed_from


def change_user_state(user_inst=None, user_state=None):
    """

    :param user_inst:
    :param user_state:
    :return:
    """
    changed = user_inst.state_from_dict(user_state)
    # mark active
    user_inst.mark_activity()
    if changed:
        channels = user_inst.get_channels()
        for channel in [c for c in channels if c.notify_state]:
            channel.send_user_state(user_inst, changed)
    return changed


def disconnect(conn_id):
    """

    :param conn_id:
    :return:
    """
    server_state = get_state()
    conn = server_state.connections.get(conn_id)
    if conn is not None:
        conn.mark_for_gc()
        return True
    return False


def set_channel_config(channel_configs):
    """

    :param channel_configs:
    :return:
    """
    server_state = get_state()
    with server_state.lock:
        for channel_name, config in channel_configs.items():
            if not server_state.channels.get(channel_name):
                channel = Channel(
                    channel_name, channel_config=channel_configs.get(channel_name)
                )
                server_state.channels[channel_name] = channel
            else:
                channel = server_state.channels[channel_name]
                channel.reconfigure_from_dict(channel_configs.get(channel_name))


def pass_message(msg, stats):
    """

    :param msg:
    :param stats:
    :return:
    """
    server_state = get_state()
    msg["catchup"] = False
    msg["edited"] = None
    msg["type"] = "message"

    total_sent = 0
    stats["total_unique_messages"] += 1
    if msg.get("channel"):
        channel_inst = server_state.channels.get(msg["channel"])
        if channel_inst:
            total_sent += channel_inst.add_message(
                msg, pm_users=msg["pm_users"], exclude_users=msg["exclude_users"]
            )
    elif msg["pm_users"]:
        # if pm then iterate over all users and notify about new message!
        for username in msg["pm_users"]:
            user_inst = server_state.users.get(username)
            if user_inst:
                total_sent += user_inst.add_message(msg)
    stats["total_messages"] += total_sent


def edit_message(msg):
    """

    :param msg:
    :return:
    """
    server_state = get_state()
    if msg.get("channel"):
        channel_inst = server_state.channels.get(msg["channel"])
        if channel_inst:
            channel_inst.alter_message(msg)
    elif msg["pm_users"]:
        # if pm then iterate over all users and notify about new message!
        for username in msg["pm_users"]:
            user_inst = server_state.users.get(username)
            if user_inst:
                user_inst.alter_message(msg)


def delete_message(msg):
    """

    :param msg:
    :return:
    """
    server_state = get_state()
    if msg.get("channel"):
        channel_inst = server_state.channels.get(msg["channel"])
        if channel_inst:
            channel_inst.delete_message(msg)
    elif msg["pm_users"]:
        # if pm then iterate over all users and notify about new message!
        for username in msg["pm_users"]:
            user_inst = server_state.users.get(username)
            if user_inst:
                user_inst.delete_message(msg)



from gevent import monkey

monkey.patch_all()

import uuid
import pytest
import mock
from datetime import datetime
from pyramid import testing
from channelstream.server_state import get_state


@pytest.fixture
def test_uuids():
    return [
        uuid.UUID("12345678-1234-5678-1234-567812345678"),
        uuid.UUID("22345678-1234-5678-1234-567812345678"),
        uuid.UUID("32345678-1234-5678-1234-567812345678"),
        uuid.UUID("42345678-1234-5678-1234-567812345678"),
        uuid.UUID("52345678-1234-5678-1234-567812345678"),
        uuid.UUID("62345678-1234-5678-1234-567812345678"),
        uuid.UUID("72345678-1234-5678-1234-567812345678"),
        uuid.UUID("82345678-1234-5678-1234-567812345678"),
        uuid.UUID("92345678-1234-5678-1234-567812345678"),
    ]


@pytest.fixture
def cleanup_globals():
    server_state = get_state()
    server_state.channels = {}
    server_state.connections = {}
    server_state.users = {}
    server_state.stats = {
        "total_messages": 0,
        "total_unique_messages": 0,
        "started_on": datetime.utcnow(),
    }


@pytest.fixture
def pyramid_config():
    # from pyramid.request import Request
    # request = Request.blank('/', base_url='http://foo.com')
    config = testing.setUp(settings={})
    settings = config.get_settings()
    return config, settings


@pytest.fixture
def dummy_request():
    app_request = testing.DummyRequest()
    app_request.handle_cors = mock.Mock()
    return app_request



import logging

from itsdangerous import TimestampSigner
from pyramid.security import Allow, Everyone, ALL_PERMISSIONS, authenticated_userid

log = logging.getLogger(__name__)


class RequestBasicChallenge(Exception):
    pass


def is_allowed_ip(addr, config):
    if "0.0.0.0" in config["allow_posting_from"]:
        return True
    return addr in config["allow_posting_from"]


class APIFactory(object):
    def __init__(self, request):
        self.__acl__ = []
        config = request.registry.settings
        req_url_secret = request.params.get("secret")
        req_secret = request.headers.get("x-channelstream-secret", req_url_secret)

        addr = request.environ["REMOTE_ADDR"]
        if not is_allowed_ip(addr, config):
            log.warning("IP: {} is not whitelisted".format(addr))
            return

        if req_secret:
            max_age = 60 if config["validate_requests"] else None
            signer = TimestampSigner(config["secret"])
            signer.unsign(req_secret, max_age=max_age)
        else:
            return
        self.__acl__ = [(Allow, Everyone, ALL_PERMISSIONS)]


class BasicAuthFactory(object):
    def __init__(self, request):
        self.__acl__ = []
        user = authenticated_userid(request)
        if not user:
            raise RequestBasicChallenge()
        self.__acl__ = [(Allow, Everyone, ALL_PERMISSIONS)]



from gevent import monkey

monkey.patch_all()

import copy
import logging
import argparse
import sys

from six.moves import configparser

from gevent.server import StreamServer
from pyramid.settings import asbool

import channelstream.wsgi_app as pyramid_app
import channelstream
from channelstream.gc import gc_conns_forever, gc_users_forever
from channelstream.policy_server import client_handle
from channelstream.ws_app import ChatApplicationSocket

from ws4py.server.geventserver import WSGIServer
from ws4py.server.wsgiutils import WebSocketWSGIApplication

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class RoutingApplication(object):
    def __init__(self, config):
        self.ws_app = WebSocketWSGIApplication(handler_cls=ChatApplicationSocket)
        self.wsgi_app = pyramid_app.make_app(config)

    def __call__(self, environ, start_response):
        if environ["PATH_INFO"] == "/ws":
            environ["ws4py.app"] = self
            return self.ws_app(environ, start_response)

        return self.wsgi_app(environ, start_response)


SHARED_DEFAULTS = {
    "secret": "secret",
    "admin_user": "admin",
    "admin_secret": "admin_secret",
    "gc_conns_after": 30,
    "gc_channels_after": 3600 * 72,
    "wake_connections_after": 5,
    "allow_posting_from": "127.0.0.1",
    "port": 8000,
    "host": "0.0.0.0",
    "debug": False,
    "log_level": "INFO",
    "demo": False,
    "allow_cors": "",
    "validate_requests": True,
}


def cli_start():
    if sys.version_info.major < 3 or (
        sys.version_info.major <= 3 and sys.version_info.minor < 6
    ):
        logging.warning(
            "\n---\n Version 0.6.9 is the last version to support Python older than 3.6\n---\n"
        )

    config = copy.deepcopy(SHARED_DEFAULTS)

    parser = argparse.ArgumentParser(add_help=True)
    parser.add_argument(
        "-i", "--ini", dest="ini", help="Config file path", default=None
    )
    parser.add_argument(
        "-s", "--secret", dest="secret", help="Secret used to secure your API requests"
    )
    parser.add_argument(
        "-u", "--admin_username", dest="admin_user", help="Administrator username"
    )
    parser.add_argument(
        "-a",
        "--admin_secret",
        dest="admin_secret",
        help="Secret used to secure your admin panel",
    )
    parser.add_argument(
        "-host", "--host", dest="host", help="Host ip on which the server listens to"
    )
    parser.add_argument(
        "-p",
        "--port",
        type=int,
        dest="port",
        help="Port on which the server listens to",
    )
    parser.add_argument("-d", "--debug", dest="debug", help="Does nothing for now")
    parser.add_argument(
        "-l", "--log-level", dest="log_level", help="Does nothing for now"
    )
    parser.add_argument("-e", "--demo", dest="demo", help="Does nothing, BW.compat")
    parser.add_argument(
        "-x",
        "--allowed_post_ip",
        dest="allow_posting_from",
        help="comma separated list of ip's " "that can post to server",
    )
    parser.add_argument(
        "-c",
        "--allow_cors",
        dest="allow_cors",
        help="comma separated list of domains's " "that can connect to server",
    )
    parser.add_argument(
        "--validate-requests",
        dest="validate_requests",
        help="Enable timestamp check on signed requests",
    )
    args = parser.parse_args()

    parameters = (
        "debug",
        "log_level",
        "port",
        "host",
        "secret",
        "admin_user",
        "admin_secret",
        "allow_posting_from",
        "allow_cors",
        "validate_requests",
    )

    if args.ini:
        parser = configparser.ConfigParser()
        parser.read(args.ini)
        settings = dict(parser.items("channelstream"))
        for key in parameters:
            try:
                config[key] = settings[key]
            except KeyError:
                pass
    else:
        for key in parameters:
            conf_value = getattr(args, key)
            if conf_value:
                config[key] = conf_value

    # convert types
    config["debug"] = asbool(config["debug"])
    config["port"] = int(config["port"])
    config["validate_requests"] = asbool(config["validate_requests"])

    for key in ["allow_posting_from", "allow_cors"]:
        if not config[key]:
            continue
        try:
            listed = [ip.strip() for ip in config[key].split(",")]
            config[key] = listed
        except ValueError:
            pass

    log_level = getattr(logging, config.get("log_level", "INFO").upper())
    logging.basicConfig(level=log_level)
    log.info("Starting channelstream {}".format(channelstream.__version__))
    url = "http://{}:{}".format(config["host"], config["port"])

    log.info("Starting flash policy server on port 10843")
    gc_conns_forever()
    gc_users_forever()
    server = StreamServer(("0.0.0.0", 10843), client_handle)
    server.start()
    log.info("Serving on {}".format(url))
    log.info("Admin interface available on {}/admin".format(url))
    if config["secret"] == "secret":
        log.warning("Using default secret! Remember to set that for production.")
    if config["admin_secret"] == "admin_secret":
        log.warning("Using default admin secret! Remember to set that for production.")

    server = WSGIServer(
        (config["host"], config["port"]),
        RoutingApplication(config),
        log=logging.getLogger("channelstream.WSGIServer"),
    )
    server.serve_forever()



import collections
import uuid

import marshmallow
from apispec.ext.marshmallow.openapi import OpenAPIConverter
from marshmallow import fields, ValidationError
from marshmallow.base import FieldABC

from channelstream.server_state import get_state

converter = OpenAPIConverter("2.0.0")


MSG_EDITABLE_KEYS = ("uuid", "timestamp", "user", "message", "edited")


try:
    base_types = (unicode, basestring)
except NameError:
    base_types = (str,)


def gen_uuid():
    return uuid.uuid4()


def validate_connection_id(conn_id):
    server_state = get_state()
    if conn_id not in server_state.connections:
        raise marshmallow.ValidationError("Unknown connection")


def validate_username(username):
    server_state = get_state()
    if username not in server_state.users:
        raise marshmallow.ValidationError("Unknown user")


def add_missing_fields(data, original, form_fields):
    for key, val in original.items():
        if key not in form_fields:
            data[key] = val
    return data


@converter.map_to_openapi_type("object", "object")
class UserStateDictField(fields.Dict):
    default_error_messages = {
        "invalid_value": "'{dict_key}' key is not type of string, integer, boolean or float.",
        "invalid": "Not a valid mapping type.",
    }

    def _deserialize(self, value, attr, data):
        value = super(UserStateDictField, self)._deserialize(value, attr, data)
        for k in value.keys():
            if (
                not isinstance(value[k], (str, bytes))
                and not isinstance(value[k], float)
                and not isinstance(value[k], int)
                and not isinstance(value[k], bool)
                and value[k] is not None
            ):
                self.fail("invalid_value", dict_key=k)
        return value


@converter.map_to_openapi_type("string", "string")
class UserStateField(fields.Field):
    default_error_messages = {
        "invalid_value": "Value is not type of string, integer, boolean or float."
    }

    def _deserialize(self, value, attr, data):
        if (
            not isinstance(value, base_types)
            and not isinstance(value, float)
            and not isinstance(value, int)
            and not isinstance(value, bool)
            and value is not None
        ):
            self.fail("invalid_value")
        return value


# backported from marshmallow 3.x
@converter.map_to_openapi_type("object", "object")
class BackportedDict(fields.Field):
    """A dict field. Supports dicts and dict-like objects. Optionally composed
    with another `Field` class or instance.

    Example: ::

        numbers = fields.Dict(values=fields.Float(), keys=fields.Str())

    :param Field values: A field class or instance for dict values.
    :param Field keys: A field class or instance for dict keys.
    :param kwargs: The same keyword arguments that :class:`Field` receives.

    .. note::
        When the structure of nested data is not known, you may omit the
        `values` and `keys` arguments to prevent content validation.

    .. versionadded:: 2.1.0
    """

    default_error_messages = {"invalid": "Not a valid mapping type."}

    def __init__(self, values=None, keys=None, **kwargs):
        super(BackportedDict, self).__init__(**kwargs)
        if values is None:
            self.value_container = None
        elif isinstance(values, type):
            if not issubclass(values, FieldABC):
                raise ValueError(
                    '"values" must be a subclass of ' "marshmallow.base.FieldABC"
                )
            self.value_container = values()
        else:
            if not isinstance(values, FieldABC):
                raise ValueError(
                    '"values" must be of type ' "marshmallow.base.FieldABC"
                )
            self.value_container = values
        if keys is None:
            self.key_container = None
        elif isinstance(keys, type):
            if not issubclass(keys, FieldABC):
                raise ValueError(
                    '"keys" must be a subclass of ' "marshmallow.base.FieldABC"
                )
            self.key_container = keys()
        else:
            if not isinstance(keys, FieldABC):
                raise ValueError('"keys" must be of type ' "marshmallow.base.FieldABC")
            self.key_container = keys

    def _add_to_schema(self, field_name, schema):
        super(BackportedDict, self)._add_to_schema(field_name, schema)
        if self.value_container:
            self.value_container.parent = self
            self.value_container.name = field_name
        if self.key_container:
            self.key_container.parent = self
            self.key_container.name = field_name

    def _serialize(self, value, attr, obj):
        if value is None:
            return None
        if not self.value_container and not self.key_container:
            return value
        if isinstance(value, collections.Mapping):
            values = value.values()
            if self.value_container:
                values = [
                    self.value_container._serialize(item, attr, obj) for item in values
                ]
            keys = value.keys()
            if self.key_container:
                keys = [self.key_container._serialize(key, attr, obj) for key in keys]
            return dict(zip(keys, values))
        self.fail("invalid")

    def _deserialize(self, value, attr, data):
        if not isinstance(value, collections.Mapping):
            self.fail("invalid")
        if not self.value_container and not self.key_container:
            return value

        errors = collections.defaultdict(dict)
        values = list(value.values())
        keys = list(value.keys())
        if self.key_container:
            for idx, key in enumerate(keys):
                try:
                    keys[idx] = self.key_container.deserialize(key)
                except ValidationError as e:
                    errors[key]["key"] = e.messages
        if self.value_container:
            for idx, item in enumerate(values):
                try:
                    values[idx] = self.value_container.deserialize(item)
                except ValidationError as e:
                    values[idx] = e.data
                    key = keys[idx]
                    errors[key]["value"] = e.messages
        result = dict(zip(keys, values))

        if errors:
            raise ValidationError(errors, data=result)

        return result


class ChannelstreamSchema(marshmallow.Schema):
    class Meta:
        strict = True
        ordered = True



from six.moves.urllib.parse import parse_qs
from ws4py.websocket import WebSocket

from channelstream import utils
from channelstream.server_state import get_state


class ChatApplicationSocket(WebSocket):
    def __init__(self, *args, **kwargs):
        super(ChatApplicationSocket, self).__init__(*args, **kwargs)
        self.qs = None
        self.conn_id = None

    def opened(self):
        server_state = get_state()
        self.qs = parse_qs(self.environ["QUERY_STRING"])
        self.conn_id = utils.uuid_from_string(self.qs.get("conn_id")[0])
        if self.conn_id not in server_state.connections:
            # close connection instantly if user played with id
            self.close()
        else:
            # attach a socket to connection
            connection = server_state.connections[self.conn_id]
            connection.socket = self
            connection.deliver_catchup_messages()

    def received_message(self, m):
        server_state = get_state()
        # this is to allow client heartbeats
        if self.conn_id in server_state.connections:
            connection = server_state.connections[self.conn_id]
            connection.mark_activity()
            user = server_state.users.get(connection.username)
            if user:
                user.mark_activity()

    def closed(self, code, reason=""):
        server_state = get_state()
        self.environ.pop("ws4py.app")
        found_conn = self.conn_id in server_state.connections
        if hasattr(self, "conn_id") and found_conn:
            connection = server_state.connections[self.conn_id]
            connection.mark_for_gc()



import copy
import logging
import uuid
from datetime import datetime

import six

from channelstream.server_state import get_state
from channelstream.utils import process_catchup
from channelstream.validation import MSG_EDITABLE_KEYS

log = logging.getLogger(__name__)


class User(object):
    """ represents a unique user of the system """

    def __init__(self, username):
        self.uuid = uuid.uuid4()
        self.username = username
        self.state = {}
        self.state_public_keys = []
        self.connections = []  # holds ids of connections
        # store frames for fetching when connection is established
        # those frames will store private messages
        self.frames = []
        self.last_active = None
        self.mark_activity()

    def mark_activity(self):
        self.last_active = datetime.utcnow()

    def __repr__(self):
        return "<User:%s, connections:%s>" % (self.username, len(self.connections))

    def add_frame(self, frame):
        self.frames.append((datetime.utcnow(), frame))
        self.frames = self.frames[-50:]

    def get_catchup_frames(self, newer_than):
        return [process_catchup(f[1]) for f in self.frames if f[0] > newer_than]

    def add_connection(self, connection):
        """
        creates a new connection for user
        """
        if connection not in self.connections:
            self.connections.append(connection)
        # mark active
        self.mark_activity()
        return connection

    def add_message(self, message):
        """
        Send a message to all connections of this user
        """
        # mark active
        message = copy.deepcopy(message)
        message.pop("no_history", None)
        message.pop("pm_users", None)
        message.pop("exclude_users", None)
        self.add_frame(message)
        self.mark_activity()
        for connection in self.connections:
            connection.add_message(message)
        return len(self.connections)

    def state_from_dict(self, state_dict):
        changed = []
        if isinstance(state_dict, dict):
            for k, v in six.iteritems(state_dict):
                if self.state.get(k) != v:
                    self.state[k] = v
                    changed.append({"key": k, "value": v})
        return changed

    @property
    def public_state(self):
        return {k: v for k, v in self.state.items() if k in self.state_public_keys}

    def get_info(self, include_connections=False):
        info = {"state": self.public_state, "user": self.username, "uuid": self.uuid}
        if include_connections:
            info["connections"] = [c.id for c in self.connections]
        return info

    def get_channels(self):
        server_state = get_state()
        channels = []
        for channel in six.itervalues(server_state.channels):
            if channel.connections.get(self.username):
                channels.append(channel)
        return channels

    def alter_message(self, to_edit):
        # normally tried to get channel and user from history
        for f, msg in self.frames:
            if msg["uuid"] == to_edit["uuid"] and msg["type"] == "message":
                msg.update(
                    {k: v for k, v in six.iteritems(to_edit) if k in MSG_EDITABLE_KEYS}
                )
                break
        altered = copy.deepcopy(to_edit)
        altered["type"] = "message:edit"
        self.add_message(altered)

    def delete_message(self, to_delete):
        # normally tried to get channel and user from history
        for i, frame in enumerate(self.frames):
            msg = frame[1]
            if msg["uuid"] == to_delete["uuid"] and msg["type"] == "message":
                self.frames.pop(i)
                break

        deleted = copy.deepcopy(to_delete)
        deleted["type"] = "message:delete"
        self.add_message(deleted)

    def __json__(self, request=None):
        return self.get_info()



from pyramid.events import NewRequest
from pyramid.events import subscriber


@subscriber(NewRequest)
def use_prefix(event):
    # rewrite prefix if X-Script-Name header is present
    script_name = event.request.headers.get("X_SCRIPT_NAME", "")
    if script_name:
        event.request.environ["SCRIPT_NAME"] = script_name
        path_info = event.request.environ["PATH_INFO"]
        if path_info.startswith(script_name):
            event.request.environ["PATH_INFO"] = path_info[len(script_name) :]

    # make sure http/https can be controlled
    scheme = event.request.headers.get("X_SCHEME", "")
    scheme = event.request.headers.get("X_URL_SCHEME", scheme)
    if scheme:
        event.request.environ["wsgi.url_scheme"] = scheme



import logging
from datetime import datetime, timedelta

import gevent
import six

from channelstream import patched_json as json
from channelstream.server_state import get_state

log = logging.getLogger(__name__)


class Connection(object):
    """ Represents a client connection"""

    def __init__(self, username, conn_id):
        self.username = username  # hold user id/name of connection
        self.last_active = None
        self.socket = None
        self.queue = None
        self.id = conn_id
        self.mark_activity()
        gevent.spawn_later(5, self.heartbeat)

    def __repr__(self):
        return "<Connection: id:%s, owner:%s>" % (self.id, self.username)

    def mark_activity(self):
        self.last_active = datetime.utcnow()

    def add_message(self, message=None):
        server_state = get_state()
        """ Sends the message to the client connection """
        # handle websockets
        if self.socket and self.socket.terminated:
            self.mark_for_gc()
        elif self.socket and not self.socket.terminated:
            try:
                # payload needs to be converted to JSON now as it gets
                # piped to client
                self.socket.send(json.dumps([message] if message else []))
                self.mark_activity()
                server_state.users[self.username].mark_activity()
            except Exception as exc:
                log.info(exc)
                self.mark_for_gc()
        elif self.queue:
            # handle long polling
            # payload will be converted to JSON in WSGI response
            self.queue.put([message] if message else [])

    def mark_for_gc(self):
        # set last active time for connection 1 hour in past for GC
        self.last_active -= timedelta(days=60)

    def heartbeat(self):
        if self.socket or self.queue:
            try:
                self.add_message()
                gevent.spawn_later(5, self.heartbeat)
            except Exception:
                self.mark_for_gc()
                if self.socket:
                    self.socket.close()

    def get_catchup_messages(self):
        server_state = get_state()
        messages = []
        # return catchup messages for channels
        for channel in self.channels:
            channel_inst = server_state.channels[channel]
            messages.extend(
                channel_inst.get_catchup_frames(self.last_active, self.username)
            )
        # and users
        messages.extend(
            server_state.users[self.username].get_catchup_frames(self.last_active)
        )
        return messages

    def deliver_catchup_messages(self):
        [self.add_message(m) for m in self.get_catchup_messages()]

    @property
    def channels(self):
        """
        Return list of channels names connection belongs to
        :return:
        """
        server_state = get_state()
        found_channels = []
        for channel in six.itervalues(server_state.channels):
            user_conns = channel.connections.get(self.username) or []
            if self in user_conns:
                found_channels.append(channel.name)
        return sorted(found_channels)

    def __json__(self):
        return self.id
